import logging

import numpy as np
import treefiles as tf

from gcma.io.baseio import IOEncoder
from gcma.io.feature import TFeatures, Features
from gcma.io.param import TParams, Params


class Hist:
    def __init__(
        self, params: TParams, features: TFeatures, fvalue: float, it: int = None
    ):
        """
        Stores a generation's best results

        :param params: best parameters of the generation `it`
        :param features: best features of the generation `it`
        :param fvalue: value of the global error
        :param it: iteration
        """
        self.params = params
        self.features = features
        self.fvalue = fvalue
        self.iteration = it

    def to_dict(self) -> dict:
        """
        Dump to dict
        """
        return {
            "iteration": self.iteration,
            "params": self.params,
            "features": self.features,
            "fvalue": self.fvalue,
        }

    @classmethod
    def from_dict(cls, d: dict):
        """
        Load from dict
        """
        return cls(
            Params(d["params"]),
            Features(d["features"]),
            float(d["fvalue"]),
            it=int(d["iteration"]),
        )

    def __repr__(self):
        s = f"Iteration {self.iteration}, fvalue = {self.fvalue}\n"
        s += f"Params:\n{Params(self.params).table()}\n"
        s += f"Features:\n{Features(self.features).table()}"
        return s


class HistoryEncoder(IOEncoder):
    def default(self, obj):
        if isinstance(obj, Hist):
            return obj.to_dict()
        return super().default(obj)


class History(list):
    @property
    def best_i(self):
        """
        Return the best iteration
        """
        fvalues = np.array([x.fvalue for x in self])
        return np.argmin(fvalues)

    @property
    def best(self):
        """
        Return the best history among all generations
        """
        return self[self.best_i]

    @property
    def last(self):
        """
        Return the last generation's history
        """
        return self[-1]

    def add(self, params: TParams, features: TFeatures, fvalue: float):
        """
        Add a history record

        :param params: best parameters of the generation `it`
        :param features: best features of the generation `it`
        :param fvalue: value of the global error
        """
        self.append(Hist(params, features, fvalue, it=len(self)))

    def dump(self, fname: str):
        """
        Export history to json file

        :param fname: filename
        """
        tf.dump_json(fname, self, cls=HistoryEncoder)

    @classmethod
    def load(cls, fname: str):
        """
        Load history from json file

        :param fname: filename
        """
        l = [Hist.from_dict(x) for x in tf.load_json(fname)]
        return History(l)


log = logging.getLogger(__name__)
